/* DockTile.dock.java
* Docktile Library for Processing
* exhibits a GraphicsContext which draws into the app's docktile in the dock
* again, most of this code is taken from the late JDIC project @ jdic.dev.java.net
* created by derDoc (derDoc@gmx.de) @ 12.4.2008

 * Copyright (C) 2008 derDoc. All rights reserved. Use is
 * subject to license terms.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.
*/



package DockTile;

import java.awt.*;
import java.awt.image.*;
import processing.core.*;


public class dock extends PGraphics3D {

	//the colorspace name that the NSBitmapImageRep will use
  	private static final String colorSpaceName = "NSCalibratedRGBColorSpace"; //if doesn't work try NSDeviceRGBColorSpace or NSCalibratedRGBColorSpace
  
  	//static string to be used inside a processing sketch's createGraphics method
	public static final String DOCK = "DockTile.dock";


  public void setParent(PApplet p) {  // ignore
    super.setParent(p);
	//if not supported, or not right Java version, exit
//	if ( (p != null && p.javaVersion < 1.6f) || (!SystemTray.isSupported()) ){
//		throw new RuntimeException("SystemTray not supported");
//	}

  }

	public void setSize(int w, int h) {  // ignore
		super.setSize(w,h);
		initNative();
	}
	
	//called by P5 internally
	public dock() { }
	
	/*
	public dock(int w, int h) {
		this(w,h,null);
	} 
	
	public dock(int w, int h, PApplet p) {
		super(w,h,p);
		initNative();
	} 
	*/
	
//	-------------- Processing functions ---------------
	

	/*
 	 * Class:     DockTile.dock
 	 * Method:    allocate
 	 * called by processing when the size of this graphics instance has changed
 	 * used to allocate structures, and buffers
 	*/
	protected void allocate(){
		//call super to set up everything
		super.allocate();
		setupProcessing();
	}
	
	
	//shuts down = cleanup
	public void dispose(){
		shutdownNative();
	}


	/*
 	 * Class:     DockTile.dock
 	 * Method:    displayable
 	 * called by processing to find out if this should be displayed in the main window
 	*/
  	public boolean displayable() {
    	return false;  // just in case someone wants to use this as the main drawing surface
  	}
	

	/*
 	 * Class:     DockTile.dock
 	 * Method:    endDraw
 	 * called by processing when a drawing loop has finished
 	 * just pass the pixels to the native method
 	*/
	public void endDraw(){
		super.endDraw();
		setPixelsNative(this.pixels);
	}


//	---------------------   own functions ---------------------------------------------------
	private native 	void initNative();
	private native	void setupNative(int length, int imageWidth, int imageHeight,
                                        int bitsPerSample, int samplesPerPixel, boolean hasAlpha, boolean isPlanar,
                                        String colorSpaceName, int bytesPerRow, int bitsPerPixel, boolean inProcessing);
	private native 	void setImageNative(byte[] rasterData);
	private native 	void setPixelsNative(int[] pixels);
	private native 	void shutdownNative();
	private native 	void setAlphaNative(float alpha);
	private native 	void setBadgeNative(String label);
	
	
	/* static constructor
	* gets called when the class is loaded by the classloader
	* just load the DockTile JNI native lib
	*/
	static {    	
        System.loadLibrary("DockTile");
	}


	/*
 	 * Class:     DockTile.dock
 	 * Method:    setup
 	 * Param: Image - the image to use for getting its infos, for buffers etc.
 	 * call to setup parameters of the image to pass to the native code
 	*/
	private void setup(Image im){
		
		Graphics2D g;
		int i,n;

        if (im != null)
        {
        	
            BufferedImage tmpImage = getBufferedImage(im);
            tmpImage.flush();

            WritableRaster wr = tmpImage.getRaster();
            DataBuffer db = wr.getDataBuffer();


            byte[] pixels = ((DataBufferByte) db).getData(); //this is stored as ABGR, so we need to swap the B & R bytes to get RGB, but do that later
            
            n = pixels.length;
			
            int imageWidth = wr.getWidth();
            int imageHeight = wr.getHeight();
            int samplesPerPixel = wr.getNumBands();
            int bitsPerSample = 8;
            boolean hasAlpha = true;
            boolean isPlanar = false;

            int bytesPerRow = imageWidth * samplesPerPixel;
            int bitsPerPixel = samplesPerPixel * bitsPerSample;

            setupNative(n, imageWidth, imageHeight,
                            bitsPerSample, samplesPerPixel, hasAlpha, isPlanar,
                            colorSpaceName, bytesPerRow, bitsPerPixel, false);

		}
	}

	
	/*
 	 * Class:     DockTile.dock
 	 * Method:    setupProcessing
 	 * call to setup standard parameters of the image to pass to the native code when inside processing
 	*/
	private void setupProcessing(){
            setupNative(pixels.length*4, width, height,
                            8, 4, true, false,
                            colorSpaceName, width * 4, 32, true);
	}


	/*
 	 * Class:   DockTile.dock
 	 * Method:  setImage
 	 * Param:	Image
 	 * gets the pixles of the supplied image and passes them to the native code as byte[]
 	*/
	private void setImage(Image im){
			int i,n;
	
	        BufferedImage tmpImage = getBufferedImage(im);
            
            tmpImage.flush();

            WritableRaster wr = tmpImage.getRaster();
			DataBuffer db = wr.getDataBuffer();

			byte[] pix = ((DataBufferByte) db).getData(); //this is stored as ABGR, so we need to swap the B & G bytes
            n = pix.length;
            
            /*
             * now done inside the native C code (faster - uses int[])
            for (i = 0; i < n; i += 4)
            {
                byte swap = pix[i + 1];
                pix[i + 1] = pix[i + 3];
                pix[i + 3] = swap;
            }
            */
            
			//now just set the image into the docktile, all structures all already in place
			setImageNative(pix);
	
	}
	
	
	/* Method: setPixels
	 * takes the pixels of this processing instance, switches B&G and passes them to the native code
	 * the switching is done now entirely in C, so no need for this function anymore
	private void setPixels(){
	//still need to swap B&G in the buffer, first lets get a copy
//		int[] xcopy = this.pixels.clone();
		//now go over it and change
	//	int n = xcopy.length;
		
		for (int i = 0; i < n; i ++)
        {
        	xcopy[i] = ( ((xcopy[i] & 0x000000FF) << 16) | ((xcopy[i] & 0xFF00FF00)) | ((xcopy[i] & 0x00FF0000) >> 16) );
		}
		setPixelsNative(this.pixels);
	}
	*/
	
	//sets the windows alpha value
	public void setAlpha(float alpha){
		setAlphaNative(alpha);
	}
	
	//sets the badge label of the app's dock tile
	public void setBadge(String badge){
		setBadgeNative(badge);
	}
	

	//returns a java.awt.BufferedImage of the supplied image, because this is in most cases apple.awt.Image - which is totally undocumented
	private static BufferedImage getBufferedImage(Image img){
   
   		if( img == null ) return null;
   		int w = img.getWidth(null);
   		int h = img.getHeight(null);

	   	// draw original image to thumbnail image object and
   		// scale it to the new size on-the-fly
   		BufferedImage bufimg = new BufferedImage(w, h, BufferedImage.TYPE_4BYTE_ABGR);
   		Graphics2D g2 = bufimg.createGraphics();
   		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
   		g2.drawImage(img, 0, 0, w, h, null);
   		g2.dispose();
   		return bufimg;
	}


	
	//-----------------   just for testing, to be able to start the .jar directly (needs middle.gif and loading.gif from Processing in the same dir  -------------------------------------------
	public static void main(String args[]){
		System.out.println("In main, making new dock instance");
		new dock(100,100);
	}
	
	//this creates a new dock instance, when started directly from the jar
	public dock(int w, int h){
		System.out.println("In dock(w,h), calling initfromJar");
		initFromJar(w,h);
	}

	private void initFromJar(int w, int h){
		System.out.println("In initfromjar, starting...");
	
		//create the frame
		Frame f = new Frame("DockTile.dock test");
		f.setSize(500,500);
		//f.pack();
		f.setVisible(true);
		
		System.out.println("In initfromjar, created frame, loading icons...");
		//load the icon
		Image b = Toolkit.getDefaultToolkit().getImage("middle.gif");
		Image c = Toolkit.getDefaultToolkit().getImage("loading.gif");
		
		while (b.getWidth(null) < 0 && b.getHeight(null) < 0 ){
			try{ Thread.sleep(10); } catch(Exception e) {}
		}
		
		System.out.println("In initfromjar, done loading icons, calling initNative...");
		//init the NSWindow
		initNative();
		//set up mem and structs
		System.out.println("In initfromjar, calling setup with b...");
		setup(b);
		
		System.out.println("In initfromjar, setting image to b...");
		//and set its icons to something
		setImage(b);
		//and its badge and alpha
		setAlpha(0.75f);
		setBadge("test");
		try{ Thread.sleep(1000); } catch(Exception e) {}
		while (c.getWidth(null) < 0 && c.getHeight(null) < 0 ){
			try{ Thread.sleep(10); } catch(Exception e) {}
		}
		
		System.out.println("In initfromjar, setup with c and setting image to c...");
		setup(c);
		setImage(c);
		try{ Thread.sleep(1000); } catch(Exception e) { e.printStackTrace(); }
		
		System.out.println("In initfromjar, calling shutdown native...");
		shutdownNative();
		System.exit(0);
	}
	
	
	
	
    /*
 	 * Class:     DockTile.dock
 	 * Method:    temp
 	 * takes a java.awt.Image and spits out a bunch of info about it, like the number of samples, pixels/sample etc.
 	 * taken from the JDIC project (jdic.dev.java.net)
 	*/
	private void temp(Image im)
    {

        //here is the data I need to supply to create an NSBitmapImageRep in the native Mac code:
        /*

			bmrep = [bmrep initWithBitmapDataPlanes:&image->buffer_ptr
			pixelsWide:(GLint) image->buffer_size.width
			pixelsHigh:(GLint) image->buffer_size.height
			bitsPerSample:8
			samplesPerPixel:4
			hasAlpha:YES
			isPlanar:NO
			colorSpaceName:NSCalibratedRGBColorSpace
			bytesPerRow:image->buffer_size.width*4
			bitsPerPixel:32];

        */
        Graphics2D g;
        Image iconImage = null;
        if (im != null)
        {
            if (iconImage == null)
            {
                iconImage = getBufferedImage(im);

                BufferedImage bi = (BufferedImage) iconImage; //for casting
                System.out.println("type of BufferedImage is " + iconImage.getClass().getName());
                System.out.println("width of image is " + bi.getWidth());
                System.out.println("height of image is " + bi.getHeight());

                Raster r = bi.getData();
                System.out.println("bi.getDate() : type of the Raster is " + r.getClass().getName());
                DataBuffer db = r.getDataBuffer();
                System.out.println("type of the DataBuffer is " + db.getClass().getName());
                byte pixels[] = ((DataBufferByte) db).getData();
                System.out.println("size of pixel array is " + pixels.length);

                System.out.println("");
                WritableRaster wr = bi.getRaster();
                System.out.println("bi.getRaster() : type of the Raster is " + wr.getClass().getName());
                DataBuffer db2 = wr.getDataBuffer();
                System.out.println("type of DataBuffer is " + db2.getClass().getName());

                pixels = ((DataBufferByte) db2).getData();
                System.out.println("size of pixel array is " + pixels.length);
                System.out.println("raster bouds is " + wr.getBounds());
                System.out.println("raster width=" + wr.getWidth() + ", height=" + wr.getHeight());
                System.out.println("raster minX=" + wr.getMinX() + ", minY=" + wr.getMinY());
                System.out.println("raster number bands (samples per pixel)=" + wr.getNumBands());
                System.out.println("raster num data elements =" + wr.getNumDataElements());
                System.out.println("");
                int imageWidth = wr.getWidth();
                int imageHeight = wr.getHeight();
                int samplesPerPixel = wr.getNumBands();
                int bitsPerSample = 8;
                boolean hasAlpha = true;
                boolean isPlanar = false;
                String colorSpaceName = "NSCalibratedRGBColorSpace"; //if doesn't work try NSDeviceRGBColorSpace
                int bytesPerRow = imageWidth * samplesPerPixel;
                int bitsPerPixel = samplesPerPixel * bitsPerSample;

                SampleModel sm = bi.getSampleModel();
                System.out.println("type of the SampleModel is " + sm.getClass().getName());
                int[] sampleSize = sm.getSampleSize();
                System.out.println("length of sample size array is " + sampleSize.length);
                System.out.println("samples are :");
                for (int i = 0; i < sampleSize.length; i++)
                {
                    int i1 = sampleSize[i];
                    System.out.print("" + i + ": " + i1 + ", ");
                }
                System.out.println("");
                if (sm.getClass() == SinglePixelPackedSampleModel.class)
                {
                    System.out.println("scanline stride is " + ((SinglePixelPackedSampleModel) sm).getScanlineStride());
                }
            }
        }
    }
}
