/* Native JNI library for drawing into the Dock Tile of a Java Application
*  Most of this code is taken from the native part of the late JDIC project @ http://jdic.dev.java.net/
   Written by derDoc (derDoc@gmx.de) @ 12.4.2008
   
 * Copyright (C) 2008 derDoc. All rights reserved. Use is
 * subject to license terms.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA.

*/

#import <Foundation/Foundation.h>
#import <Cocoa/Cocoa.h>
#import <inttypes.h>
#import <JavaVM/jawt_md.h>

#import "DockTile_dock.h"


//-- forward functions defs -----

 NSString *ConvertToNSString(JNIEnv *env, jstring str);


// ------------ vars -----------

NSAutoreleasePool *pool;
NSWindow * nsw;
unsigned char *bufPtr, *dataPtr;
NSBitmapImageRep *bmrep;
NSImage *bmi;

NSDockTile *tile;
NSImageView *nsiv, * nsiv2;
jbyte *carr;
jint *car2;
jbyte temp[64];
NSSize nss;
       

int i,n;
jint pixBitmapFormat = (NSAlphaNonpremultipliedBitmapFormat); //| NSAlphaFirstBitmapFormat 
jint ImBitmapFormat  = (NSAlphaNonpremultipliedBitmapFormat | NSAlphaFirstBitmapFormat);
NSString *colorSpace;
BOOL alpha, plan;


// ---------- func defs --------------


/*
 * Class:     DockTile_dock
 * Method:    initNative
 * Signature: ()V
 * just get the docktile and make a new NSImageView, and set it as the view of the DockTile
 */
JNIEXPORT void JNICALL Java_DockTile_dock_initNative
  (JNIEnv *env, jobject this){
  		pool = [[NSAutoreleasePool alloc] init];

		//try to get the current running application
		NSApplication *nsa = [NSApplication sharedApplication];
 		nsw = [nsa mainWindow];
		
  		//get the app dockTile
  		tile = [NSApp dockTile];
//		NSLog(@"obj-C(init): tile size: %f, %f.", [tile size].width, [tile size].height);
  		
  		//make the ImageView
  		NSRect frame = NSMakeRect(0, 0, tile.size.width, tile.size.height);
		nsiv = [[NSImageView alloc] initWithFrame: frame];
 		[nsiv autorelease];
  		[nsiv setEditable:false];
  		[nsiv setAllowsCutCopyPaste:false];
  		[nsiv setAnimates:true];
  	[nsiv setImageScaling: NSScaleToFit];		//set if you want a smaller image to be scaled up to 128,128
  		
  		[tile setContentView:nsiv];

}

/*
 * Class:     DockTile_dock
 * Method:    setupNative
 * Signature: (IIIIZZLjava/lang/String;II)V
 * is called to set up the NSImage, the NSBitmapImageRep and various buffers, needs all kinds of params of the pic (like its size etc)
 * also needs to know if its called from within Processing, since the ordering samples in a pixel is different
 */
JNIEXPORT void JNICALL Java_DockTile_dock_setupNative
  (JNIEnv *env, jobject this, jint len, jint imageWidth, jint imageHeight,
                            jint bitsPerSample, jint samplesPerPixel, jboolean hasAlpha, jboolean isPlanar,
                            jstring colorSpaceName, jint bytesPerRow, jint bitsPerPixel, jboolean inProcessing){
  
		pool = [[NSAutoreleasePool alloc] init];
        jbyte temp[len];
        carr = temp;
		
        if (carr == NULL)
        {
            return; // exception occurred
        }
		
        bufPtr = (unsigned char *)carr;
  
        alpha = (hasAlpha == 0 ? NO : YES);
        plan = (isPlanar == 0 ? NO : YES);
        colorSpace = ConvertToNSString(env, colorSpaceName);


//  		NSLog(@"obj-C(setup): called with l=%i, w=%i, h=%i, BPS=%i, SPP=%i, alpha=%i, plan=%i, CSN=%p, BPR=%i, BPP=%i, inProc=%i", 
 //	 		len, imageWidth, imageHeight, bitsPerSample, samplesPerPixel, hasAlpha, isPlanar, colorSpace, bytesPerRow, bitsPerPixel, inProcessing);
		
		nss.height=imageHeight;
		nss.width=imageWidth;
		bmi = [[NSImage alloc] initWithSize:nss];
		[bmi autorelease];

        n = (bytesPerRow *imageHeight);

        bmrep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL
				pixelsWide:imageWidth
        		pixelsHigh:imageHeight
        		bitsPerSample:bitsPerSample
        		samplesPerPixel:samplesPerPixel
        		hasAlpha:alpha
        		isPlanar:plan
				colorSpaceName:colorSpace
        		bitmapFormat: (inProcessing == 0 ? ImBitmapFormat:pixBitmapFormat)
        		bytesPerRow:bytesPerRow
        		bitsPerPixel:bitsPerPixel
        ];
        
        dataPtr = [bmrep bitmapData];
		[bmrep autorelease];
        
}

/*
 * Class:     DockTile_dock
 * Method:    setImageNative
 * Signature: ([B)V
 * call after calling setupNative to set the image into the tile
 * expects rasterdata of the image, i.e. the pixels, as byte[]
 * switches B&G in the buffer
 */
JNIEXPORT void JNICALL Java_DockTile_dock_setImageNative
  (JNIEnv *env, jobject this, jbyteArray rasterData){

//		pool = [[NSAutoreleasePool alloc] init];

		jint len = (*env)->GetArrayLength(env, rasterData);
     	jbyte temp[len];
        int intlen = len/4;
        
        jint * t;
        carr = temp;
		t = carr;
        
		(*env)->GetByteArrayRegion(env, rasterData, 0, len, carr);

		//and switch as int, which is faster than doing it in byte in java
		//switch B&G in the supplied ABGR to make it an ARGB
		for (car2 = t; car2 < (t+intlen); car2++)
			(*car2) = ( (((*car2) & 0x0000ff00) << 16) | (((*car2) & 0x00ff00ff)) | (((*car2) & 0xff000000) >> 16) );

		
        //copy the pixel data into the buffer manged by the NSBitmapImageRep
		//but use GC aware memcpy instead of sloppy original version
		objc_memmove_collectable(dataPtr, bufPtr, n);
		//and set the new representation into the image
		[bmi removeRepresentation:[bmi bestRepresentationForDevice:nil]];
		[bmi addRepresentation:bmrep];
		

        if ([bmi isValid])
        {
//            NSLog(@"obj-C(setIm): image is valid and can be rendered, w/ size [%f,%f]", bmi.size.width, bmi.size.height);
            
			//set image for dock tile
			[nsiv setImage:bmi];
			//and tell the tile to redraw
            [tile display];
        }
        else
        {
            NSLog(@"obj-C(setIm): Image is NOT valid");
        }
		
}





/*
 * Class:     DockTile_dock
 * Method:    setPixelsNative
 * Signature: ([I)V
 * Function for direct calling when used within Processing, it allows to pass the pixels directly form P's pixels array (its int[])
 * w/out taking the longer route of taking P's image->bufferedImage->writableRaster->DataBuffer etc.
 */
JNIEXPORT void JNICALL Java_DockTile_dock_setPixelsNative
  (JNIEnv * env, jobject this, jintArray data){
  
  		jint l = (*env)->GetArrayLength(env, data);
       	jint arr[l];
        bufPtr = (unsigned char *)arr;
  
		(*env)->GetIntArrayRegion(env, data, 0, l, arr);

//		NSLog(@"setPixles: switching B&G in %p for length of int=%i", car2, l);
		//switch B&G in the supplied ABGR to make it an ARGB
		//note that the order of fields is different than from an awt.BufferedImage, alpha is not in the MSBs (!alphafirst)
		for (car2 = arr; car2 < (arr+l); car2++)
			(*car2) = ( (((*car2) & 0x000000ff) << 16) | (((*car2) & 0xff00ff00)) | (((*car2) & 0x00ff0000) >> 16) );


        dataPtr = [bmrep bitmapData];
		//move the data into the imrep
		objc_memmove_collectable(dataPtr, bufPtr, n);
		//and readd it to the im
		[bmi removeRepresentation:[bmi bestRepresentationForDevice:nil]];
		[bmi addRepresentation:bmrep];


        if ([bmi isValid])
        {
//            NSLog(@"obj-C(setIm): image is valid and can be rendered, w/ size [%f,%f]", bmi.size.width, bmi.size.height);
            
			//set image for dock tile
			[nsiv setImage:bmi];
            [tile display];
        }
        else
        {
            NSLog(@"obj-C(setIm): Image is NOT valid");
        }
  
}




/*
 * Class:     DockTile_dock
 * Method:    shutdownNative
 * Signature: ()V
 * shutdown and release images
 */
JNIEXPORT void JNICALL Java_DockTile_dock_shutdownNative
  (JNIEnv *env, jobject this){

      [nsiv setImage:nil];
      [tile setContentView:nil];
	  [pool release];

}

 /*
 * Class:     DockTile_dock
 * Method:    setAlphaNative
 * Signature: (F)V
 * sets the alpha of the main window, does not work from processing
 */
JNIEXPORT void JNICALL Java_DockTile_dock_setAlphaNative
  (JNIEnv * env, jobject obj, jfloat alpha){
  
   		//set its alpha
 		[nsw setAlphaValue:alpha];

  }

/*
 * Class:     DockTile_dock
 * Method:    setBadgeNative
 * Signature: (Ljava/lang/String;)V
 * sets a lavel to the dock tile
 */
JNIEXPORT void JNICALL Java_DockTile_dock_setBadgeNative
  (JNIEnv * env, jobject this, jstring badge){
  
  		NSString * label = ConvertToNSString(env, badge);
  		[tile setBadgeLabel:label];
  
  }




/* From JDIC
 * ConvertToNSString
 *
 * given a non-null jstring argument, return the equivalent NSString representation. The object is autoreleased.
 *
 * This function returns NULL if the argument is NULL, or if the NSString couldn't be created. Requires the JNIEnv
 * to be passed as the first argument
 *
 */
NSString *ConvertToNSString(JNIEnv *env, jstring str)
{
    if (str == NULL)
    {
        return NULL;
    }

    const jchar *chars = (*env)->GetStringChars(env, str, NULL);
    NSString *myNSString =
        [NSString stringWithCharacters:(UniChar *)chars  length:(*env)->GetStringLength(env, str)];
    (*env)->ReleaseStringChars(env, str, chars);

    return myNSString;
}